/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#include <cassert>
#include <chrono>  // std::chrono::milliseconds
#include <numeric> // std::accumulate
#include <thread>  // std::this_thread::sleep_for
#include <vector>

#include "aidge/operator/Conv.hpp"
#include "aidge/utils/Types.h"

#include "aidge/backend/cpu/operator/AddImpl.hpp"
#include "aidge/backend/cpu/operator/AddImpl_forward_kernels.hpp"

//////////////////////////////////
// AddImpl_cpu<1>
//////////////////////////////////

Aidge::NbElts_t Aidge::AddImpl_cpu<1>::getNbRequiredProtected(const Aidge::IOIndex_t /*inputIdx*/) const {
    // this implementation can be in-place
    return 0;
}

void Aidge::AddImpl_cpu<1>::forward() {
    assert(mOp.getInput(0) && "missing input #0");

    // Find the correct kernel type
    auto kernelFunc = Registrar<AddImplForward_cpu<1>>::create({
        mOp.getInput(0)->dataType(),
        mOp.getOutput(0)->dataType()});

    // Call kernel
    kernelFunc(std::static_pointer_cast<Tensor>(mOp.getInput(0))->size(),
        mOp.getInput(0)->getImpl()->rawPtr(),
        mOp.getOutput(0)->getImpl()->rawPtr());
}


//////////////////////////////////
// AddImpl_cpu<2>
//////////////////////////////////

Aidge::NbElts_t Aidge::AddImpl_cpu<2>::getNbRequiredProtected(const Aidge::IOIndex_t /*inputIdx*/) const {
    // this implementation of add can be in-place
    return 0;
}

void Aidge::AddImpl_cpu<2>::forward() {
    assert(mOp.getInput(0) && "missing input #0");
    assert(mOp.getInput(1) && "missing input #1");

    // Find the correct kernel type
    auto kernelFunc = Registrar<AddImplForward_cpu<2>>::create({
        mOp.getInput(0)->dataType(),
        mOp.getInput(1)->dataType(),
        mOp.getOutput(0)->dataType()});

    // Call kernel
    kernelFunc(std::static_pointer_cast<Tensor>(mOp.getInput(0))->size(),
        mOp.getInput(0)->getImpl()->rawPtr(),
        mOp.getInput(1)->getImpl()->rawPtr(),
        mOp.getOutput(0)->getImpl()->rawPtr());
}


//////////////////////////////////
// AddImpl_cpu<3>
//////////////////////////////////

Aidge::NbElts_t Aidge::AddImpl_cpu<3>::getNbRequiredProtected(const Aidge::IOIndex_t /*inputIdx*/) const {
    // this implementation of add can be in-place
    return 0;
}

void Aidge::AddImpl_cpu<3>::forward() {
    assert(mOp.getInput(0) && "missing input #0");
    assert(mOp.getInput(1) && "missing input #1");
    assert(mOp.getInput(2) && "missing input #2");

    // Find the correct kernel type
    auto kernelFunc = Registrar<AddImplForward_cpu<3>>::create({
        mOp.getInput(0)->dataType(),
        mOp.getInput(1)->dataType(),
        mOp.getInput(2)->dataType(),
        mOp.getOutput(0)->dataType()});

    // Call kernel
    kernelFunc(std::static_pointer_cast<Tensor>(mOp.getInput(0))->size(),
        mOp.getInput(0)->getImpl()->rawPtr(),
        mOp.getInput(1)->getImpl()->rawPtr(),
        mOp.getInput(2)->getImpl()->rawPtr(),
        mOp.getOutput(0)->getImpl()->rawPtr());
}
